/**
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.deephacks.tools4j.config.internal.core;

import com.google.common.base.Optional;
import com.google.common.base.Strings;
import com.google.common.collect.Lists;
import org.deephacks.tools4j.config.internal.core.query.ConfigIndex;
import org.deephacks.tools4j.config.internal.core.query.ConfigIndexedCollection;
import org.deephacks.tools4j.config.internal.core.xml.XmlBeanManager;
import org.deephacks.tools4j.config.internal.core.xml.XmlSchemaManager;
import org.deephacks.tools4j.config.model.Bean;
import org.deephacks.tools4j.config.model.Bean.BeanId;
import org.deephacks.tools4j.config.model.Lookup;
import org.deephacks.tools4j.config.model.Schema;
import org.deephacks.tools4j.config.model.SystemProperties;
import org.deephacks.tools4j.config.spi.BeanManager;
import org.deephacks.tools4j.config.spi.CacheManager;
import org.deephacks.tools4j.config.spi.SchemaManager;

import javax.enterprise.inject.Instance;
import javax.inject.Inject;
import javax.inject.Singleton;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import static org.deephacks.tools4j.config.model.Events.CFG101_SCHEMA_NOT_EXIST;

/**
 * ConfigCore implements functionality that is common to both admin and runtime context.
 */
@Singleton
public final class ConfigCore {

    @Inject
    private Instance<SchemaManager> schemaManagerInstances;

    @Inject
    private Instance<BeanManager> beanManagerInstances;

    @Inject
    private Instance<CacheManager> cacheManagerInstances;

    private static Optional<CacheManager> cacheManager;

    @Inject
    private SystemProperties properties;

    private static final String CONFIG_QUERY_FEATURE_ENABLED_PROPERTY = "config.query.enabled";
    private boolean configQueryFeatureEnabled = false;
    /** schemaName -> index */
    private static final HashMap<String, ConfigIndex> configIndexes = new HashMap<>();

    /** schemaName -> indexed collection */
    private static final HashMap<String, ConfigIndexedCollection> indexCollections = new HashMap<>();

    public void setSchema(Map<String, Schema> schemas, Collection<Bean> beans) {
        for (Bean bean : beans) {
            setSchema(bean, schemas);
        }
    }

    public void setSchema(Bean b, Map<String, Schema> schemas) {
        Schema s = schemas.get(b.getId().getSchemaName());
        if (s == null) {
            throw CFG101_SCHEMA_NOT_EXIST(b.getId().getSchemaName());
        }
        b.set(s);
        for (BeanId id : b.getReferences()) {
            Bean ref = id.getBean();
            if (ref != null && ref.getSchema() == null) {
                setSchema(ref, schemas);
            }
        }
    }

    public void setSchema(Map<String, Schema> schemas, Map<BeanId, Bean> beans) {
        for (Bean b : beans.values()) {
            setSchema(b, schemas);
        }
    }

    public BeanManager lookupBeanManager() {

        BeanManager manager = lookup(BeanManager.class, beanManagerInstances, "config.beanmanager");
        if (manager != null) {
            return manager;
        }
        /**
         * return as default, zero-configuration required by users
         */
        return new XmlBeanManager();
    }

    public SchemaManager lookupSchemaManager() {
        SchemaManager manager = lookup(SchemaManager.class, schemaManagerInstances,
                "config.schemamanager");
        if (manager != null) {
            return manager;
        }
        /**
         * return as default, zero-configuration required by users
         */
        return new XmlSchemaManager();
    }

    public Optional<CacheManager> lookupCacheManager() {
        CacheManager manager = lookup(CacheManager.class, cacheManagerInstances,
                "config.cachemanager");
        if(!configQueryFeatureEnabled) {
            return Optional.absent();
        }
        if (manager != null) {
            cacheManager = Optional.of(manager);
        } else {
            cacheManager = Optional.absent();
        }
        return cacheManager;
    }

    private <T> T lookup(Class<T> clazz, Instance<T> instances, String property) {
        if (properties == null) {
            properties = SystemProperties.createDefault();
            if(!Strings.isNullOrEmpty(properties.get(CONFIG_QUERY_FEATURE_ENABLED_PROPERTY))){
                configQueryFeatureEnabled = true;
            }
        }
        String prefered = properties.get(property);
        if (!Strings.isNullOrEmpty(prefered)) {
            T instance = getPreferedInstance(instances, prefered);
            if (instance != null) {
                return instance;
            }
            return clazz.cast(Reflections.newInstance(prefered));
        }
        T instance = getPreferedInstance(instances, null);
        if (instance != null) {
            return instance;
        }

        Collection<T> managers = Lookup.get().lookupAll(clazz);
        if (managers.size() == 1) {
            return managers.iterator().next();
        } else {
            return Lookup.get().lookup(clazz);
        }
    }

    private <T> T getPreferedInstance(Instance<T> instances, String prefered) {
        LinkedList<T> preferedInstances = new LinkedList<>();
        if (instances == null) {
            return null;
        }
        List<T> list = Lists.newArrayList(instances);
        for (T instance : list) {
            if (Strings.isNullOrEmpty(prefered)) {
                String packageName = instance.getClass().getPackage().getName();
                if (packageName.contains("internal.core.xml")) {
                    preferedInstances.addLast(instance);
                } else {
                    preferedInstances.addFirst(instance);
                }
            } else if (instance.getClass().getName().equals(prefered)) {
                return instance;
            }
        }
        return preferedInstances.getFirst();
    }

    public void putIndex(Schema schema) {
        if(!configQueryFeatureEnabled) {
            return;
        }
        if(configIndexes.get(schema.getName()) != null) {
            return;
        }
        cacheManager.get().registerSchema(schema);
        ConfigIndex index = new ConfigIndex(schema);
        configIndexes.put(schema.getName(), index);
        indexCollections.put(schema.getName(), new ConfigIndexedCollection(index));
    }

    public void removeIndex(Schema schema) {
        if(!configQueryFeatureEnabled) {
            return;
        }
        String schemaName = schema.getName();
        indexCollections.remove(schemaName);
        configIndexes.remove(schemaName);
        if(cacheManager.isPresent()) {
            cacheManager.get().clear(schemaName);
        }
    }

    public void cache(Collection<Bean> beans) {
        for (Bean bean : beans) {
            cache(bean);
        }
    }

    public void cache(Bean bean) {
        if(!configQueryFeatureEnabled) {
            return;
        }
        if(!cacheManager.isPresent()) {
            throw new IllegalArgumentException("A cache manager is needed to to queries.");
        }
        ConfigIndexedCollection col = indexCollections.get(bean.getId().getSchemaName());
        col.add(bean);
        cacheManager.get().put(bean);
    }

    public ConfigIndexedCollection get(Schema schema) {
        if(!configQueryFeatureEnabled) {
            throw new IllegalArgumentException("Config query is not enabled.");
        }
        return indexCollections.get(schema.getName());
    }

    public void cacheRemove(BeanId beanId) {
        if(!configQueryFeatureEnabled) {
            return;
        }
        if(!cacheManager.isPresent()) {
            return;
        }
        cacheManager.get().remove(beanId);
        ConfigIndexedCollection col = indexCollections.get(beanId.getSchemaName());
        col.remove(beanId);
    }

    public void cacheRemove(String schemaName, Collection<String> instances) {
        for (String instanceId : instances) {
            cacheRemove(BeanId.create(instanceId, schemaName));
        }
    }
}
