/**
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.deephacks.tools4j.config.internal.core.runtime;

import com.google.common.base.Optional;
import org.deephacks.tools4j.config.Config;
import org.deephacks.tools4j.config.RuntimeContext;
import org.deephacks.tools4j.config.internal.core.ConfigCore;
import org.deephacks.tools4j.config.internal.core.query.ConfigIndexedCollection;
import org.deephacks.tools4j.config.internal.core.query.ConfigQuery;
import org.deephacks.tools4j.config.model.AbortRuntimeException;
import org.deephacks.tools4j.config.model.Bean;
import org.deephacks.tools4j.config.model.Bean.BeanId;
import org.deephacks.tools4j.config.model.Events;
import org.deephacks.tools4j.config.model.Schema;
import org.deephacks.tools4j.config.model.Schema.SchemaPropertyRef;
import org.deephacks.tools4j.config.spi.BeanManager;
import org.deephacks.tools4j.config.spi.CacheManager;
import org.deephacks.tools4j.config.spi.Conversion;
import org.deephacks.tools4j.config.spi.SchemaManager;

import javax.inject.Inject;
import javax.inject.Singleton;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import static org.deephacks.tools4j.config.model.Events.CFG102_NOT_CONFIGURABLE;
import static org.deephacks.tools4j.config.model.Events.CFG303;

/**
 * RuntimeCoreContext is responsible for separating the admin, runtime and spi
 * context so that no dependencies (compile nor runtime) exist between them.
 */
@Singleton
public final class RuntimeCoreContext extends RuntimeContext {
    private Conversion conversion;
    private SchemaManager schemaManager;
    private BeanManager beanManager;
    private Optional<CacheManager> cacheManager;

    @Inject
    private ConfigCore core;

    @Override
    public void register(Class<?>... configurable) {
        doLookup();
        for (Class<?> clazz : configurable) {
            Schema schema = conversion.convert(clazz, Schema.class);
            schemaManager.regsiterSchema(schema);
            core.putIndex(schema);
        }
    }

    @Override
    public void unregister(Class<?>... configurable) {
        doLookup();
        for (Class<?> clazz : configurable) {
            Schema schema = conversion.convert(clazz, Schema.class);
            schemaManager.removeSchema(schema.getName());
            core.removeIndex(schema);
        }
    }

    @Override
    public void registerDefault(Object... instances) {
        doLookup();
        for (Object instance : instances) {
            Bean bean = conversion.convert(instance, Bean.class);
            bean.setDefault();
            try {
                beanManager.create(bean);
            } catch (AbortRuntimeException e) {
                // ignore if bean already exist
                if (e.getEvent().getCode() != CFG303) {
                    throw e;
                }
            }
        }
    }

    @Override
    public <T> T singleton(Class<T> configurable) {
        doLookup();
        Schema schema = conversion.convert(configurable, Schema.class);
        BeanId singleton = getSingletonId(schema, configurable);
        Map<String, Schema> schemas = schemaManager.getSchemas();
        Bean bean = beanManager.getEager(singleton);
        core.setSchema(bean, schemas);
        setSingletonReferences(bean, schemas);
        T object = conversion.convert(bean, configurable);
        core.cache(bean);
        return object;
    }

    @Override
    public <T> List<T> all(Class<T> clazz) {
        doLookup();
        Schema s = getSchema(clazz);
        Map<String, Schema> schemas = schemaManager.getSchemas();
        Map<BeanId, Bean> beans = beanManager.list(s.getName());
        core.setSchema(schemas, beans);
        for (Bean bean : beans.values()) {
            setSingletonReferences(bean, schemas);
            core.cache(bean);
        }
        ArrayList<T> objects = new ArrayList<>();
        for(T object : conversion.convert(beans.values(), clazz)) {
            objects.add(object);

        }
        return objects;
    }

    @Override
    public <T> T get(String id, Class<T> clazz) {
        doLookup();
        Schema s = getSchema(clazz);
        Map<String, Schema> schemas = schemaManager.getSchemas();
        BeanId beanId = BeanId.create(id, s.getName());
        Bean bean = beanManager.getEager(beanId);
        if (bean == null) {
            throw Events.CFG304_BEAN_DOESNT_EXIST(beanId);
        }
        core.setSchema(bean, schemas);
        setSingletonReferences(bean, schemas);
        T object = conversion.convert(bean, clazz);
        core.cache(bean);
        return object;
    }

    @Override
    public <T> org.deephacks.tools4j.config.model.query.ConfigQuery newQuery(Class<T> configurable) {
        doLookup();
        Schema schema = getSchema(configurable);
        ConfigIndexedCollection collection = core.get(schema);
        if(collection == null) {
            throw Events.CFG101_SCHEMA_NOT_EXIST(configurable.getName());
        }
        if(!cacheManager.isPresent()) {
            throw new IllegalStateException("Queries are not possible without a cache manager.");
        }
        return new ConfigQuery<T>(collection, cacheManager.get());
    }

    private BeanId getSingletonId(Schema s, Class<?> configurable) {
        return BeanId.createSingleton(s.getName());
    }

    private void setSingletonReferences(Bean bean, Map<String, Schema> schemas) {
        Schema s = bean.getSchema();
        for (SchemaPropertyRef ref : s.get(SchemaPropertyRef.class)) {
            if (ref.isSingleton()) {
                Schema singletonSchema = schemas.get(ref.getSchemaName());
                Bean singleton = beanManager.getSingleton(ref.getSchemaName());
                singleton.set(singletonSchema);
                BeanId singletonId = singleton.getId();
                singletonId.setBean(singleton);
                // recursive call.
                setSingletonReferences(singleton, schemas);
                bean.setReference(ref.getName(), singletonId);
            }
        }
    }

    public void doLookup() {
        // core would already be injected in a cdi environment
        if (core == null) {
            core = new ConfigCore();
        }
        if (conversion == null) {
            conversion = Conversion.get();
            conversion.register(new ObjectToBeanConverter());
            conversion.register(new ClassToSchemaConverter());
            conversion.register(new FieldToSchemaPropertyConverter());
            conversion.register(new BeanToObjectConverter());

        }
        if (beanManager == null) {
            beanManager = core.lookupBeanManager();
        }
        if (schemaManager == null) {
            schemaManager = core.lookupSchemaManager();
        }
        if (cacheManager == null) {
            cacheManager = core.lookupCacheManager();
        }

    }

    private Schema getSchema(Class<?> clazz) {
        ClassIntrospector introspector = new ClassIntrospector(clazz);
        Config config = introspector.getAnnotation(Config.class);
        if (config == null) {
            throw CFG102_NOT_CONFIGURABLE(clazz);
        }
        String schemaName = config.name();
        if (schemaName == null || "".equals(schemaName)) {
            schemaName = clazz.getName();
        }
        return schemaManager.getSchema(schemaName);
    }
}
